@@ -30,12 +30,6 @@ class ImageLoadingError(ValueError):
 
     """
 
-    @classmethod
-    def from_exception(cls, exception):
-        name = type(exception).__name__
-        value = str(exception)
-        return cls(f'{name}: {value}' if value else name)
-
 
 class RasterImage:
     def __init__(self, pillow_image, image_id, image_data, filename=None,
@@ -257,7 +251,7 @@ class LazyLocalImage(pydyf.Object):
 class SVGImage:
     def __init__(self, tree, base_url, url_fetcher, context):
         font_config = context.font_config if context else None
-        self._svg = SVG(tree, base_url, font_config)
+        self._svg = SVG(tree, base_url, font_config, url_fetcher)
         self._base_url = base_url
         self._url_fetcher = url_fetcher
         self._context = context
@@ -284,7 +278,7 @@ class SVGImage:
         try:
             self._svg.draw(
                 stream, concrete_width, concrete_height, self._base_url,
-                self._url_fetcher, self._context)
+                self._context)
         except BaseException as exception:
             LOGGER.error('Failed to render SVG image %s', self._base_url)
             LOGGER.debug('Error while rendering SVG image:', exc_info=exception)
@@ -297,43 +291,40 @@ def get_image_from_uri(cache, url_fetcher, options, url, forced_mime_type=None,
         return cache[url]
 
     try:
-        with fetch(url_fetcher, url) as result:
-            if 'string' in result:
-                string = result['string']
-            else:
-                string = result['file_obj'].read()
-            mime_type = forced_mime_type or result['mime_type']
+        with fetch(url_fetcher, url) as response:
+            bytestring = response.read()
+            mime_type = forced_mime_type or response.content_type
 
         image = None
         svg_exceptions = []
         # Try to rely on given mimetype for SVG
         if mime_type == 'image/svg+xml':
             try:
-                tree = ElementTree.fromstring(string)
+                tree = ElementTree.fromstring(bytestring)
                 image = SVGImage(tree, url, url_fetcher, context)
             except Exception as svg_exception:
                 svg_exceptions.append(svg_exception)
         # Try pillow for raster images, or for failing SVG
         if image is None:
             try:
-                pillow_image = Image.open(BytesIO(string))
+                pillow_image = Image.open(BytesIO(bytestring))
             except Exception as raster_exception:
                 if mime_type == 'image/svg+xml':
                     # Tried SVGImage then Pillow for a SVG, abort
-                    raise ImageLoadingError.from_exception(svg_exceptions[0])
+                    raise ImageLoadingError from svg_exceptions[0]
                 try:
                     # Last chance, try SVG
-                    tree = ElementTree.fromstring(string)
+                    tree = ElementTree.fromstring(bytestring)
                     image = SVGImage(tree, url, url_fetcher, context)
                 except Exception:
                     # Tried Pillow then SVGImage for a raster, abort
-                    raise ImageLoadingError.from_exception(raster_exception)
+                    raise ImageLoadingError from raster_exception
             else:
                 # Store image id to enable cache in Stream.add_image
                 image_id = md5(url.encode(), usedforsecurity=False).hexdigest()
-                path = result.get('path')
                 image = RasterImage(
-                    pillow_image, image_id, string, path, cache, orientation, options)
+                    pillow_image, image_id, bytestring, response.path, cache,
+                    orientation, options)
 
     except (URLFetchingError, ImageLoadingError) as exception:
         LOGGER.error('Failed to load image at %r: %s', url, exception)
@@ -655,7 +646,7 @@ class LinearGradient(Gradient):
                 positions.append(positions[-1] + step)
                 last += step * stop_length
 
-            # Add colors before last step
+            # Add colors before first step
             while first > 0:
                 step = next(previous_steps)
                 colors.insert(0, next(previous_colors))
@@ -764,14 +755,15 @@ class RadialGradient(Gradient):
             center_x, center_y / scale_y, last)
 
         if self.repeating:
-            points, positions, colors = self._repeat(
-                width, height, scale_y, points, positions, colors)
+            points, positions, colors, hints = self._repeat(
+                width, height, scale_y, points, positions, colors, hints)
 
         return scale_y, 'radial', points, positions, colors, hints
 
-    def _repeat(self, width, height, scale_y, points, positions, colors):
+    def _repeat(self, width, height, scale_y, points, positions, colors, hints):
         # Keep original lists and values, they’re useful
         original_colors = colors.copy()
+        original_hints = hints.copy()
         original_positions = positions.copy()
         gradient_length = points[5] - points[2]
 
@@ -787,13 +779,14 @@ class RadialGradient(Gradient):
             # Repeat colors and extrapolate positions
             repeat = 1 + repeat_after
             colors *= repeat
+            hints = ([*hints, 1] * repeat)[:-1]
             positions = [
                 i + position for i in range(repeat) for position in positions]
             points = (*points[:5], points[5] + gradient_length * repeat_after)
 
         if points[2] == 0:
             # Inner circle has 0 radius, no need to repeat inside, return
-            return points, positions, colors
+            return points, positions, colors, hints
 
         # Find how many times we have to repeat the colors inside
         repeat_before = points[2] / gradient_length
@@ -806,6 +799,7 @@ class RadialGradient(Gradient):
         if full_repeat:
             # Repeat colors and extrapolate positions
             colors += original_colors * full_repeat
+            hints += [1, *original_hints] * full_repeat
             positions = [
                 i - full_repeat + position for i in range(full_repeat)
                 for position in original_positions] + positions
@@ -814,7 +808,7 @@ class RadialGradient(Gradient):
         partial_repeat = repeat_before - full_repeat
         if partial_repeat == 0:
             # No partial repeat, return
-            return points, positions, colors
+            return points, positions, colors, hints
 
         # Iterate through positions in reverse order, from the outer
         # circle to the original inner circle, to find positions from
@@ -828,11 +822,12 @@ class RadialGradient(Gradient):
                 # The center is a color of the gradient, truncate original
                 # colors and positions and prepend them
                 colors = original_colors[-i:] + colors
+                hints = [*original_hints[-i:], 1, *hints]
                 new_positions = [
                     position - full_repeat - 1
                     for position in original_positions[-i:]]
                 positions = new_positions + positions
-                return points, positions, colors
+                return points, positions, colors, hints
             if position < ratio:
                 # The center is between two colors of the gradient,
                 # define the center color as the average of these two
@@ -842,14 +837,14 @@ class RadialGradient(Gradient):
                 next_position = original_positions[-(i - 1)]
                 average_colors = [color, color, next_color, next_color]
                 average_positions = [position, ratio, ratio, next_position]
-                zero_color = gradient_average_color(
-                    average_colors, average_positions)
-                colors = [zero_color, *original_colors[-(i - 1):], *colors]
+                zero_color = gradient_average_color(average_colors, average_positions)
+                colors = [zero_color, *original_colors[-(i-1):], *colors]
+                hints = [1, *original_hints[-(i-1):], 1, *hints]
                 new_positions = [
                     position - 1 - full_repeat for position
                     in original_positions[-(i - 1):]]
                 positions = (ratio - 1 - full_repeat, *new_positions, *positions)
-                return points, positions, colors
+                return points, positions, colors, hints
 
     def _resolve_size(self, width, height, center_x, center_y, style):
         """Resolve circle size of the radial gradient."""
